import logging
import argparse
import os
import os.path
import subprocess
import sys
import typing

import crmsh.constants
import crmsh.parallax
import crmsh.user_of_host
import crmsh.utils


logger = logging.getLogger(__name__)


class Feature:
    _feature_registry = dict()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        Feature._feature_registry[cls.__name__.rsplit('.', 1)[-1]] = cls

    @staticmethod
    def get_feature_by_name(name: str):
        return Feature._feature_registry[name]

    def check_quick(self) -> bool:
        raise NotImplementedError

    def check_local(self, nodes: typing.Iterable[str]) -> bool:
        """Check whether the feature is functional on local node."""
        raise NotImplementedError

    def check_cluster(self, nodes: typing.Iterable[str]) -> bool:
        """Check whether the feature is functional on the cluster."""
        raise NotImplementedError

    def fix_local(self, nodes: typing.Iterable[str], ask: typing.Callable[[str], None]) -> None:
        """Fix the feature on local node.

        At least one of fix_local and fix_cluster should be implemented. If fix_local is not implemented, this method
        will be run on each node.
        """
        raise NotImplementedError

    def fix_cluster(self, nodes: typing.Iterable[str], ask: typing.Callable[[str], None]) -> None:
        """Fix the feature on the cluster.

        At least one of fix_local and fix_cluster should be implemented. If this method is not implemented, fix_local
        will be run on each node.
        """
        raise NotImplementedError


class FixFailure(Exception):
    pass


class AskDeniedByUser(Exception):
    pass


def feature_quick_check(feature: Feature):
    return feature.check_quick()


def feature_local_check(feature: Feature, nodes: typing.Iterable[str]):
    try:
        if not feature.check_quick():
            return False
    except NotImplementedError:
        pass
    return feature.check_local(nodes)


def feature_full_check(feature: Feature, nodes: typing.Iterable[str]) -> bool:
    try:
        if not feature.check_quick():
            return False
    except NotImplementedError:
        pass
    try:
        if not feature.check_local(nodes):
            return False
    except NotImplementedError:
        pass
    try:
        return feature.check_cluster(nodes)
    except NotImplementedError:
        results = crmsh.parallax.parallax_run(
            nodes,
            '/usr/bin/env python3 -m crmsh.healthcheck check-local {}'.format(
                feature.__class__.__name__.rsplit('.', 1)[-1],
            )
        )
        return all(rc == 0 for rc, _, _ in results.values())


def feature_fix(feature: Feature, nodes: typing.Iterable[str], ask: typing.Callable[[str], None]) -> None:
    try:
        return feature.fix_cluster(nodes, ask)
    except NotImplementedError:
        results = crmsh.parallax.parallax_run(
            nodes,
            '/usr/bin/env python3 -m crmsh.healthcheck fix-local {}'.format(
                feature.__class__.__name__.rsplit('.', 1)[-1],
            )
        )
        if any(rc != 0 for rc, _, _ in results.values()):
            raise FixFailure


class PasswordlessHaclusterAuthenticationFeature(Feature):
    SSH_DIR = os.path.expanduser('~hacluster/.ssh')
    KEY_TYPES = ['ed25519', 'ecdsa', 'rsa']

    def __str__(self):
        return "Configure Passwordless for hacluster"

    def check_quick(self) -> bool:
        for key_type in self.KEY_TYPES:
            try:
                os.stat('{}/id_{}'.format(self.SSH_DIR, key_type))
                os.stat('{}/id_{}.pub'.format(self.SSH_DIR, key_type))
                return True
            except FileNotFoundError:
                pass
        return False

    def check_local(self, nodes: typing.Iterable[str]) -> bool:
        try:
            for node in nodes:
                subprocess.check_call(
                    ['sudo', 'su', '-', 'hacluster', '-c', 'ssh {} hacluster@{} true'.format(crmsh.constants.SSH_OPTION, node)],
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            return True
        except subprocess.CalledProcessError:
            return False

    def fix_cluster(self, nodes: typing.Iterable[str], ask: typing.Callable[[str], None]) -> None:
        import crmsh.bootstrap  # import bootstrap lazily here to avoid circular dependency
        logger.debug("setup passwordless ssh authentication for user hacluster")
        local_node = crmsh.utils.this_node()
        remote_nodes = set(nodes)
        try:
            remote_nodes.remove(local_node)
        except KeyError:
            # bsc#1223438: local nodename not in cib
            # init_ssh_impl should work even if the local node is included in user_node_list
            # although this is not a designed usage
            logger.warning("local node %s is not found in cluster node list %s", local_node, remote_nodes)
        remote_nodes = list(remote_nodes)
        crmsh.parallax.parallax_run(
            nodes,
            'chown hacluster: ~hacluster/.ssh/authorized_keys && chmod 0600 ~hacluster/.ssh/authorized_keys',
        )
        crmsh.bootstrap.configure_ssh_key('hacluster')
        crmsh.bootstrap.swap_key_for_hacluster(remote_nodes)
        for node in remote_nodes:
            crmsh.bootstrap.change_user_shell('hacluster', node)


class PasswordlessPrimaryUserAuthenticationFeature(Feature):
    def check_quick(self) -> bool:
        local_node = crmsh.utils.this_node()
        try:
            crmsh.utils.user_of(local_node)
            return True
        except crmsh.user_of_host.UserNotFoundError:
            return False

    def check_local(self, nodes: typing.Iterable[str]) -> bool:
        try:
            for node in nodes:
                crmsh.utils.user_of(node)
        except crmsh.user_of_host.UserNotFoundError:
            return False
        try:
            crmsh.parallax.parallax_call(nodes, 'true')
            return True
        except ValueError:
            return False

    def fix_local(self, nodes: typing.Iterable[str], ask: typing.Callable[[str], None]) -> None:
        logger.warning('Passwordless ssh is not initialized. Use `crm cluster init ssh` (on one node) and `crm cluster join ssh -c <init-node>` (on other nodes) to set it up.')
        raise FixFailure

    def fix_cluster(self, nodes: typing.Iterable[str], ask: typing.Callable[[str], None]) -> None:
        return self.fix_local(nodes, ask)


def main_check_local(args) -> int:
    try:
        feature = Feature.get_feature_by_name(args.feature)()
        nodes = crmsh.utils.list_cluster_nodes(no_reg=True)
        if nodes:
            if feature_local_check(feature, nodes):
                return 0
            else:
                return 1
    except KeyError:
        logger.error('No such feature: %s.', args.feature)
    return 2


def main_fix_local(args) -> int:
    try:
        feature = Feature.get_feature_by_name(args.feature)()
        nodes = crmsh.utils.list_cluster_nodes(no_reg=True)
        if nodes:
            if args.yes:
                def ask(msg): return True
            else:
                def ask(msg): return crmsh.utils.ask('Healthcheck: fix: ' + msg, background_wait=False)
            if args.without_check or not feature_local_check(feature, nodes):
                feature.fix_local(nodes, ask)
            return 0
    except KeyError:
        logger.error('No such feature: %s.', args.feature)
    return 2


def main_fix_cluster(args) -> int:
    try:
        feature = Feature.get_feature_by_name(args.feature)()
        nodes = crmsh.utils.list_cluster_nodes(no_reg=True)
        if nodes:
            if args.yes:
                def ask(msg): return True
            else:
                def ask(msg): return crmsh.utils.ask('Healthcheck: fix: ' + msg, background_wait=False)
            if args.without_check or not feature_full_check(feature, nodes):
                feature_fix(feature, nodes, ask)
            return 0
    except KeyError:
        logger.error('No such feature: %s.', args.feature)
    return 2


def main() -> int:
    # This entrance is for internal programmatic use only.
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    check_local_parser = subparsers.add_parser('check-local')
    check_local_parser.add_argument('feature')
    check_local_parser.set_defaults(func=main_check_local)

    fix_cluster_parser = subparsers.add_parser('fix-local')
    fix_cluster_parser.add_argument('--yes', action='store_true')
    fix_cluster_parser.add_argument('--without-check', action='store_true')
    fix_cluster_parser.add_argument('feature')
    fix_cluster_parser.set_defaults(func=main_fix_local)

    fix_cluster_parser = subparsers.add_parser('fix-cluster')
    fix_cluster_parser.add_argument('--yes', action='store_true')
    fix_cluster_parser.add_argument('--without-check', action='store_true')
    fix_cluster_parser.add_argument('feature')
    fix_cluster_parser.set_defaults(func=main_fix_cluster)

    args = parser.parse_args()
    return args.func(args)


if __name__ == '__main__':
    sys.exit(main())



import unittest
from unittest import mock

import crmsh.report.sh
import crmsh.sh

import subprocess


class TestFindShell(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_cluster_shell = mock.Mock(crmsh.sh.ClusterShell)
        self.patcher_local_shell = mock.patch('crmsh.report.sh.Shell.local_shell')
        self.patcher_try_create_report_shell = mock.patch('crmsh.report.sh.Shell._try_create_report_shell')
        self.mock_local_shell = self.patcher_local_shell.start()
        self.mock_local_shell.return_value = mock.Mock(crmsh.sh.LocalShell)
        self.mock_try_create_report_shell = self.patcher_try_create_report_shell.start()

    def tearDown(self) -> None:
        self.patcher_local_shell.stop()
        self.patcher_try_create_report_shell.stop()

    def test_cluster_shell_available(self):
        self.mock_cluster_shell.can_run_as.return_value = True
        self.assertIsInstance(
                crmsh.report.sh.Shell.find_shell(self.mock_cluster_shell, 'node1', None),
                crmsh.report.sh.ClusterShellAdaptor,
        )
        self.assertIsInstance(
                crmsh.report.sh.Shell.find_shell(self.mock_cluster_shell, 'node1', 'alice'),
                crmsh.report.sh.ClusterShellAdaptor,
        )

    def test_specified_user_work(self):
        self.mock_cluster_shell.can_run_as.return_value = False
        self.mock_try_create_report_shell.return_value = mock.Mock(crmsh.report.sh.Shell)
        ret = crmsh.report.sh.Shell.find_shell(self.mock_cluster_shell, 'node1', 'alice')
        self.mock_local_shell.assert_called_once()
        self.mock_try_create_report_shell.assert_called_once_with(
            self.mock_local_shell.return_value,
            'node1', 'alice',
        )
        self.assertIs(ret, self.mock_try_create_report_shell.return_value)

class TestTryCreateReportShell(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_local_shell = mock.Mock(crmsh.sh.LocalShell)
        self.patcher_ssh_shell = mock.patch('crmsh.sh.SSHShell')
        self.mock_ssh_shell = self.patcher_ssh_shell.start().return_value

    def tearDown(self) -> None:
        self.patcher_ssh_shell.stop()

    def test_success(self):
        self.mock_ssh_shell.can_run_as.return_value = True
        self.mock_ssh_shell.subprocess_run_without_input.return_value = mock.Mock(returncode=0)
        ret = crmsh.report.sh.Shell._try_create_report_shell(self.mock_local_shell, 'node1', 'alice')
        self.mock_ssh_shell.can_run_as.assert_called_once_with('node1', 'alice')
        self.mock_ssh_shell.subprocess_run_without_input.assert_called_once_with(
            'node1', 'alice',
            'sudo true',
            start_new_session=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        self.assertIsInstance(ret, crmsh.report.sh.Shell)

    def test_failure_no_sudoer(self):
        self.mock_ssh_shell.can_run_as.return_value = True
        self.mock_ssh_shell.subprocess_run_without_input.return_value = mock.Mock(returncode=1)
        ret = crmsh.report.sh.Shell._try_create_report_shell(self.mock_local_shell, 'node1', 'alice')
        self.mock_ssh_shell.can_run_as.assert_called_once_with('node1', 'alice')
        self.mock_ssh_shell.subprocess_run_without_input.assert_called_once_with(
            'node1', 'alice',
            'sudo true',
            start_new_session=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        self.assertIsNone(ret)

    def test_failure_no_ssh(self):
        self.mock_ssh_shell.can_run_as.return_value = False
        ret = crmsh.report.sh.Shell._try_create_report_shell(self.mock_local_shell, 'node1', 'alice')
        self.mock_ssh_shell.can_run_as.assert_called_once_with('node1', 'alice')
        self.mock_ssh_shell.subprocess_run_without_input.assert_not_called()
        self.assertIsNone(ret)



import unittest
from unittest import mock

from crmsh import ui_corosync
from crmsh import corosync
from crmsh.ui_corosync import LinkArgumentParser
from crmsh.corosync import LinkManager
from crmsh.prun import prun


class TestLinkArgumentParser(unittest.TestCase):
    def test_parse_empty(self):
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, list())

    def test_invalid_link_number(self):
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, ['a0'])

    def test_no_spec(self):
        args = LinkArgumentParser().parse(True, True, ['0'])
        self.assertEqual(0, args.linknumber)
        self.assertFalse(args.nodes)
        self.assertFalse(args.options)

    def test_addr_spec(self):
        args = LinkArgumentParser().parse(True, True, ['0', 'node1=192.0.2.100', 'node2=fd00:a0::10'])
        self.assertEqual(0, args.linknumber)
        self.assertFalse(args.options)
        self.assertListEqual([('node1', '192.0.2.100'), ('node2', 'fd00:a0::10')], args.nodes)

    def test_invalid_addr_spec(self):
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, ['0', 'node1=192.0.2.300'])
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, ['0', 'node1=fd00::a0::10'])
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, ['0', 'node1=node1.example.com'])
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, ['0', 'node1=192.0.2'])

    def test_option_spec(self):
        args = LinkArgumentParser().parse(True, True, ['0', 'options', 'node1=192.0.2.100', 'node2=fd00:a0::10', 'foo='])
        self.assertEqual(0, args.linknumber)
        self.assertFalse(args.nodes)
        self.assertDictEqual({'node1': '192.0.2.100', 'node2': 'fd00:a0::10', 'foo': None}, args.options)

    def test_addrs_and_options(self):
        args = LinkArgumentParser().parse(True, True, ['0', 'node1=192.0.2.100', 'node2=fd00:a0::10', 'options', 'foo=bar=1'])
        self.assertEqual(0, args.linknumber)
        self.assertListEqual([('node1', '192.0.2.100'), ('node2', 'fd00:a0::10')], args.nodes)
        self.assertDictEqual({'foo': 'bar=1'}, args.options)

    def test_no_options(self):
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, ['0', 'options'])

    def test_garbage_inputs(self):
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, ['0', 'foo'])
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, ['0', 'node1=192.0.2.100', 'foo'])
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, ['0', 'node1=192.0.2.100', 'options', 'foo'])
        with self.assertRaises(LinkArgumentParser.SyntaxException):
            LinkArgumentParser().parse(True, True, ['0', 'node1=192.0.2.100', 'options', 'foo=bar', 'foo'])


@mock.patch('crmsh.prun.prun.prun')
@mock.patch('crmsh.utils.list_cluster_nodes')
@mock.patch('crmsh.service_manager.ServiceManager.service_is_active')
class TestCheckLinkRemovable(unittest.TestCase):
    def setUp(self):
        self.mock_lm = mock.Mock(LinkManager)

    def test_corosync_not_running(self, mock_sia, mock_lcn, mock_prun):
        mock_sia.return_value = False
        ui_corosync.Link._check_link_removable(self.mock_lm, 0)
        mock_sia.assert_called_once()
        mock_lcn.assert_not_called()
        mock_prun.assert_not_called()
        self.mock_lm.links.assert_not_called()

    def test_prun_failure(self, mock_sia, mock_lcn, mock_prun):
        mock_sia.return_value = True
        mock_lcn.return_value = ['node1', 'node2', 'node3']
        mock_prun.return_value = {
            'node2': prun.SSHError('root', 'node2', 'msg'),
        }
        with self.assertRaises(ValueError):
            ui_corosync.Link._check_link_removable(self.mock_lm, 0)
        mock_prun.assert_called_once()
        self.mock_lm.links.assert_not_called()

    def test_cmapctl_failure(self, mock_sia, mock_lcn, mock_prun):
        mock_sia.return_value = True
        mock_lcn.return_value = ['node1', 'node2', 'node3']
        mock_prun.return_value = {
            'node2': prun.ProcessResult(1, b'', b'msg'),
        }
        with self.assertRaises(ValueError):
            ui_corosync.Link._check_link_removable(self.mock_lm, 0)
        mock_prun.assert_called_once()
        self.mock_lm.links.assert_not_called()

    def test_all_links_connected(self, mock_sia, mock_lcn, mock_prun):
        mock_sia.return_value = True
        mock_lcn.return_value = ['node1', 'node2', 'node3']
        mock_prun.return_value = {
            'node1': prun.ProcessResult(
                returncode=0, stderr=b'',
                stdout=b'Local node ID 1, transport knet\n'
                       b'LINK ID 0 udp\n'
                       b'	addr	= 192.0.2.100\n'
                       b'	status:\n'
                       b'		nodeid:          1:	localhost\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	connected\n'
                       b'LINK ID 1 udp\n'
                       b'	addr	= 192.0.2.200\n'
                       b'	status:\n'
                       b'		nodeid:          1:	localhost\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	connected\n',
            ),
            'node2': prun.ProcessResult(
                returncode=0, stderr=b'',
                stdout=b'Local node ID 1, transport knet\n'
                       b'LINK ID 0 udp\n'
                       b'	addr	= 192.0.2.101\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	localhost\n'
                       b'		nodeid:          3:	connected\n'
                       b'LINK ID 1 udp\n'
                       b'	addr	= 192.0.2.201\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	localhost\n'
                       b'		nodeid:          3:	connected\n',
            ),
            'node3': prun.ProcessResult(
                returncode=0, stderr=b'',
                stdout=b'Local node ID 1, transport knet\n'
                       b'LINK ID 0 udp\n'
                       b'	addr	= 192.0.2.102\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	localhost\n'
                       b'LINK ID 1 udp\n'
                       b'	addr	= 192.0.2.202\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	localhost\n',
            ),
        }
        self.mock_lm.links.return_value = [
            corosync.Link(0, [
                corosync.LinkNode(1, 'node1', '192.0.2.100'),
                corosync.LinkNode(2, 'node2', '192.0.2.101'),
                corosync.LinkNode(3, 'node3', '192.0.2.102'),
            ]),
            corosync.Link(1, [
                corosync.LinkNode(1, 'node1', '192.0.2.100'),
                corosync.LinkNode(2, 'node2', '192.0.2.101'),
                corosync.LinkNode(3, 'node3', '192.0.2.102'),
            ]),
            0, 0, 0, 0, 0, 0,
        ]
        ui_corosync.Link._check_link_removable(self.mock_lm, 0)

    def test_one_node_pair_disconnected(self, mock_sia, mock_lcn, mock_prun):
        mock_sia.return_value = True
        mock_lcn.return_value = ['node1', 'node2', 'node3']
        mock_prun.return_value = {
            'node1': prun.ProcessResult(
                returncode=0, stderr=b'',
                stdout=b'Local node ID 1, transport knet\n'
                       b'LINK ID 0 udp\n'
                       b'	addr	= 192.0.2.100\n'
                       b'	status:\n'
                       b'		nodeid:          1:	localhost\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	connected\n'
                       b'LINK ID 1 udp\n'
                       b'	addr	= 192.0.2.200\n'
                       b'	status:\n'
                       b'		nodeid:          1:	localhost\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	connected\n',
            ),
            'node2': prun.ProcessResult(
                returncode=0, stderr=b'',
                stdout=b'Local node ID 1, transport knet\n'
                       b'LINK ID 0 udp\n'
                       b'	addr	= 192.0.2.101\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	localhost\n'
                       b'		nodeid:          3:	connected\n'
                       b'LINK ID 1 udp\n'
                       b'	addr	= 192.0.2.201\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	localhost\n'
                       b'		nodeid:          3:	disconnected\n',
            ),
            'node3': prun.ProcessResult(
                returncode=0, stderr=b'',
                stdout=b'Local node ID 1, transport knet\n'
                       b'LINK ID 0 udp\n'
                       b'	addr	= 192.0.2.102\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	localhost\n'
                       b'LINK ID 1 udp\n'
                       b'	addr	= 192.0.2.202\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	disconnected\n'
                       b'		nodeid:          3:	localhost\n',
            ),
        }
        self.mock_lm.links.return_value = [
            corosync.Link(0, [
                corosync.LinkNode(1, 'node1', '192.0.2.100'),
                corosync.LinkNode(2, 'node2', '192.0.2.101'),
                corosync.LinkNode(3, 'node3', '192.0.2.102'),
            ]),
            corosync.Link(1, [
                corosync.LinkNode(1, 'node1', '192.0.2.100'),
                corosync.LinkNode(2, 'node2', '192.0.2.101'),
                corosync.LinkNode(3, 'node3', '192.0.2.102'),
            ]),
            0, 0, 0, 0, 0, 0,
        ]
        with self.assertRaises(ValueError):
            ui_corosync.Link._check_link_removable(self.mock_lm, 0)

    def test_pair_connected(self, mock_sia, mock_lcn, mock_prun):
        mock_sia.return_value = True
        mock_lcn.return_value = ['node1', 'node2', 'node3']
        mock_prun.return_value = {
            'node1': prun.ProcessResult(
                returncode=0, stderr=b'',
                stdout=b'Local node ID 1, transport knet\n'
                       b'LINK ID 0 udp\n'
                       b'	addr	= 192.0.2.100\n'
                       b'	status:\n'
                       b'		nodeid:          1:	localhost\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	connected\n'
                       b'LINK ID 1 udp\n'
                       b'	addr	= 192.0.2.100\n'
                       b'	status:\n'
                       b'		nodeid:          1:	localhost\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	disconnected\n'
                       b'LINK ID 2 udp\n'
                       b'	addr	= 192.0.2.100\n'
                       b'	status:\n'
                       b'		nodeid:          1:	localhost\n'
                       b'		nodeid:          2:	disconnected\n'
                       b'		nodeid:          3:	disconnected\n'
                       b'LINK ID 3 udp\n'
                       b'	addr	= 192.0.2.100\n'
                       b'	status:\n'
                       b'		nodeid:          1:	localhost\n'
                       b'		nodeid:          2:	disconnected\n'
                       b'		nodeid:          3:	connected\n'
            ),
            'node2': prun.ProcessResult(
                returncode=0, stderr=b'',
                stdout=b'Local node ID 1, transport knet\n'
                       b'LINK ID 0 udp\n'
                       b'	addr	= 192.0.2.100\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	localhost\n'
                       b'		nodeid:          3:	connected\n'
                       b'LINK ID 1 udp\n'
                       b'	addr	= 192.0.2.101\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	localhost\n'
                       b'		nodeid:          3:	disconnected\n'
                       b'LINK ID 2 udp\n'
                       b'	addr	= 192.0.2.101\n'
                       b'	status:\n'
                       b'		nodeid:          1:	disconnected\n'
                       b'		nodeid:          2:	localhost\n'
                       b'		nodeid:          3:	connected\n'
                       b'LINK ID 3 udp\n'
                       b'	addr	= 192.0.2.101\n'
                       b'	status:\n'
                       b'		nodeid:          1:	disconnected\n'
                       b'		nodeid:          2:	localhost\n'
                       b'		nodeid:          3:	disconnected\n'
            ),
            'node3': prun.ProcessResult(
                returncode=0, stderr=b'',
                stdout=b'Local node ID 1, transport knet\n'
                       b'LINK ID 0 udp\n'
                       b'	addr	= 192.0.2.100\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	localhost\n'
                       b'LINK ID 1 udp\n'
                       b'	addr	= 192.0.2.102\n'
                       b'	status:\n'
                       b'		nodeid:          1:	disconnected\n'
                       b'		nodeid:          2:	disconnected\n'
                       b'		nodeid:          3:	localhost\n'
                       b'LINK ID 2 udp\n'
                       b'	addr	= 192.0.2.102\n'
                       b'	status:\n'
                       b'		nodeid:          1:	disconnected\n'
                       b'		nodeid:          2:	connected\n'
                       b'		nodeid:          3:	localhost\n'
                       b'LINK ID 3 udp\n'
                       b'	addr	= 192.0.2.102\n'
                       b'	status:\n'
                       b'		nodeid:          1:	connected\n'
                       b'		nodeid:          2:	disconnected\n'
                       b'		nodeid:          3:	localhost\n'
            ),
        }
        self.mock_lm.links.return_value = [
            corosync.Link(0, [
                corosync.LinkNode(1, 'node1', '192.0.2.100'),
                corosync.LinkNode(2, 'node2', '192.0.2.101'),
                corosync.LinkNode(3, 'node3', '192.0.2.102'),
            ]),
            corosync.Link(1, [
                corosync.LinkNode(1, 'node1', '192.0.2.100'),
                corosync.LinkNode(2, 'node2', '192.0.2.101'),
                corosync.LinkNode(3, 'node3', '192.0.2.102'),
            ]),
            corosync.Link(2, [
                corosync.LinkNode(1, 'node1', '192.0.2.100'),
                corosync.LinkNode(2, 'node2', '192.0.2.101'),
                corosync.LinkNode(3, 'node3', '192.0.2.102'),
            ]),
            corosync.Link(3, [
                corosync.LinkNode(1, 'node1', '192.0.2.100'),
                corosync.LinkNode(2, 'node2', '192.0.2.101'),
                corosync.LinkNode(3, 'node3', '192.0.2.102'),
            ]),
            0, 0, 0, 0,
        ]
        ui_corosync.Link._check_link_removable(self.mock_lm, 0)



from __future__ import unicode_literals
# Copyright (C) 2015 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.


import re

from crmsh import crm_gv
from crmsh import cibconfig


def test_digits_ident():
    g = crm_gv.gv_types["dot"]()
    cibconfig.set_graph_attrs(g, ".")

    g.new_node("1a", top_node=True)
    g.new_attr("1a", 'label', "1a")
    g.new_node("a", top_node=True)
    g.new_attr("a", 'label', "a")

    expected = [
        'fontname="Helvetica";',
        'fontsize="11";',
        'compound="true";',
        '"1a" [label="1a"];',
        'a [label="a"];',
    ]
    out = '\n'.join(g.repr()).replace('\t', '')

    for line in re.match(
            r'^digraph G {\n\n(?P<expected>.*)\n}$', out, re.M | re.S
    ).group('expected').split('\n'):
        assert line in expected
        expected.remove(line)

    assert len(expected) == 0



from .parser import lexer, LookAheadIterator
from .generator import AsciiDocGenerator

from argparse import ArgumentParser
import sys

def main():
    ap = ArgumentParser('help2adoc')
    ap.add_argument('file')
    args = ap.parse_args()
    with open(args.file, 'r') as f:
        tokens = LookAheadIterator(lexer(f))
        AsciiDocGenerator(sys.stdout.write).parse_help(tokens)
        token = tokens.lookahead()
        if token is None:
            return
        epilog_start = token.lineno
        f.seek(0)
        for i in range(epilog_start):
            next(f)
        print('....')
        for line in f:
            print(line, end='')
        print('....')


if __name__ == '__main__':
    main()
